"""
Logging module for the fritzconnection library. By default
fritzconnection will emit all messages with a log-level of INFO and
higher (WARNING, ERROR, CRITICAL).

The fritzconnection logger, defined as `fritzlogger`, is an
instance of the `logging.Logger` class from the `stdlib` and can get
accessed by:

>>> from fritzconnection.core.logger import fritzlogger

This module provides the additional functions
`activate_local_debug_mode` and `reset` (see below) to activate a debug
mode, suppressing handler-propagation, and to reset the logger to the
initial state. To activate debug logging for fritzconnection call:

>>> from fritzconnection.core.logger import activate_local_debug_mode
>>> activate_local_debug_mode(handler=logging.StreamHandler())

In this debug mode fritzconnection will additionally log the data
transfered between the library and the router and suppress
logging-propagation to the parent-handlers. Here the `StreamHandler` will
send all data to `stderr`. Because this can be a lot of data a `FileHandler`
may be a better choice:

>>> activate_local_debug_mode(handler=logging.FileHandler(<the_file>))

To deactivate this mode call `reset()`:

>>> from fritzconnection.core.logger import reset
>>> reset()

Beside this you can do everything with `fritzlogger` as with the
`logging.Logger`, because `fritzlogger` is an instance of the latter.
"""
# This module is part of the FritzConnection package.
# https://github.com/kbr/fritzconnection
# License: MIT (https://opensource.org/licenses/MIT)


import logging


fritzlogger = logging.getLogger("fritzconnection")
fritzlogger.setLevel(logging.INFO)
fritzformatter = logging.Formatter(logging.BASIC_FORMAT)


def activate_local_debug_mode(handler=None, propagate=False):
    """
    Activates all logging messages on debug level and don't propagate to
    parent-handlers. If no handler is given the NullHandler will get
    set, avoiding a call of the lastResort-handler.
    If the given handler has no formatter, the fritzformatter gets set
    (which provides the `logging.BASIC_FORMAT`).
    If propagate is True, all debug information will also get sent to
    the parent-handlers. Keep in mind, that this can be a lot of data if
    the parent-handlers are enabled for debug-level records.
    """
    if handler is None:
        handler = logging.NullHandler()
    if not handler.formatter:
        handler.setFormatter(fritzformatter)
    fritzlogger.addHandler(handler)
    fritzlogger.propagate = propagate
    fritzlogger.setLevel(logging.DEBUG)


def reset(keep_handlers=False, propagate=True):
    """
    Resets the logger to the initial state, i.e. after calling
    `activate_local_debug_mode`. All handlers will be removed, except
    `keep_handlers` is set to True.
    """
    if not keep_handlers:
        fritzlogger.handlers = []
    fritzlogger.propagate = propagate
    fritzlogger.setLevel(logging.INFO)



"""
Module to get information about WLAN devices.
"""
# This module is part of the FritzConnection package.
# https://github.com/kbr/fritzconnection
# License: MIT (https://opensource.org/licenses/MIT)
# Author: Bernd Strebel, Klaus Bremer

import io
import itertools
import random
import string

from warnings import warn

from ..core.exceptions import FritzServiceError
from .fritzbase import AbstractLibraryBase

try:
    import segno.helpers
except ImportError:
    SEGNO_INSTALLED = False
else:
    SEGNO_INSTALLED = True

# important: don't set an extension number here:
SERVICE = 'WLANConfiguration'
DEFAULT_PASSWORD_LENGTH = 12
WPA_SECURITY = 'WPA'
NO_PASS = 'nopass'


def get_beacon_security(instance, security):
    """
    Returns the beacon-security as a string based on the security
    argument. Possible return values are 'nopass' and 'WPA'. If the
    security is None or an empty string, the function tries to find the
    proper security setting ('nopass'|'WPA'). If security is neither
    None nor an empty string, the value is returned as is.

    This function is not intended to get called directly.

    .. versionadded:: 1.10
    """
    if not security:
        info = instance.get_info()
        beacontypes = set(info["NewX_AVM-DE_PossibleBeaconTypes"].split(","))
        beacontypes -= set(('None', 'OWETrans'))
        beacontype = info["NewBeaconType"]
        if beacontype in beacontypes:
            security = WPA_SECURITY
        else:
            security = NO_PASS
    return security


def get_wifi_qr_code(instance, kind='svg',
                     security=None, hidden=False,
                     scale=4):
    """
    Returns a file-like object providing a bytestring representing a
    qr-code for wlan access. `instance` is a FritzWLAN or FritzGuestWLAN
    instance. `kind` describes the type of the qr-code. Supported types
    are: 'svg', 'png' and 'pdf'. Default is 'svg'.

    This function is not intended to get called directly. Instead it is
    available as a method on FritzWLAN instances (as well as on
    subclasses like FritzGuestWLAN) if the third party package `segno`
    is installed.

    Consider `guest_wlan` is a FritzGuestWLAN instance, then the
    following code will return a file like object with the qr-code image
    data in png-format: ::

        stream = guest_wlan.get_wifi_qr_code(kind='png')

    The stream can get used anywhere, where a file like object is
    expected, i.e. writing the content to a file (Note: the suffix must
    match the kind of the qr-code format): ::

        with open('qr_code.png', 'wb') as fobj:
            fobj.write(stream.read())

    If `segno` is not installed the call will trigger an
    AttributeError when called on an instance and a NameError when
    called directly.

    .. versionadded:: 1.9.0

    The parameters `security` and `hidden` allow to forward these
    informations to the `segno` library. `security` is `None` or a
    string like `WPA`. `hidden` is a boolean value indicating the
    visibility of the network .

    .. versionadded:: 1.9.1

    If the `security` is `None` (default) the beacontype of the network
    is used to set the WLAN-Type accordingly. If the value of `security`
    is something else, this value gets used.

    `scale` defines the size of the produced qr-code. Default value is 4.

    .. versionadded:: 1.10

    """
    stream = io.BytesIO()
    security = get_beacon_security(instance, security)
    qr_code = segno.helpers.make_wifi(
        ssid=instance.ssid,
        password=instance.get_password(),
        security=security,
        hidden=hidden
    )
    qr_code.save(out=stream, kind=kind, scale=scale)
    stream.seek(0)
    return stream


def _qr_code_enabler(cls):
    """Classdecorator to inject qr-capabilities at import time."""
    if SEGNO_INSTALLED:
        cls.get_wifi_qr_code = get_wifi_qr_code
    return cls


@_qr_code_enabler
class FritzWLAN(AbstractLibraryBase):
    """
    Class to list all known wlan devices. All parameters are optional.
    If given, they have the following meaning: `fc` is an instance of
    FritzConnection, `address` the ip of the Fritz!Box, `port` the port
    to connect to, `user` the username, `password` the password,
    `timeout` a timeout as floating point number in seconds, `use_tls` a
    boolean indicating to use TLS (default False). The *service*
    parameter specifies the configuration in use. Typically this is 1
    for 2.4 GHz, 2 for 5 GHz and 3 for a guest network. This can vary
    depending on the router model and change with future standards.
    """
    def __init__(self, *args, service=1, **kwargs):
        super().__init__(*args, **kwargs)
        self.service = service

    def _action(self, actionname, **kwargs):
        service = f'{SERVICE}{self.service}'
        return self.fc.call_action(service, actionname, **kwargs)

    @property
    def host_number(self):
        """
        Number of registered wlan devices for the active
        WLANConfiguration.
        """
        result = self._action('GetTotalAssociations')
        return result['NewTotalAssociations']

    @property
    def total_host_number(self):
        """
        Total NewAssociatedDeviceIndexNumber of registered wlan devices
        for all WLANConfigurations.
        """
        total = 0
        _service = self.service
        for service in itertools.count(1):
            self.service = service
            try:
                total += self.host_number
            except FritzServiceError:
                break
        self.service = _service
        return total

    @property
    def ssid(self):
        """The WLAN SSID"""
        result = self._action('GetSSID')
        return result['NewSSID']

    @ssid.setter
    def ssid(self, value):
        self._action('SetSSID', NewSSID=value)

    @property
    def beacontype(self):
        """
        Represents the beacontype for the network as string.
        At time of writing (OS 7.29) possible values are:
        `None, 11i, WPAand11i, 11iandWPA3` for the private WiFi-security
        settings and `None, 11i, 11iandWPA3, OWETrans` for the guest
        network.
        """
        return self.get_info()['NewBeaconType']

    @property
    def channel(self):
        """The WLAN channel in use"""
        return self.channel_info()['NewChannel']

    @property
    def alternative_channels(self):
        """Alternative channels (as string)"""
        return self.channel_info()['NewPossibleChannels']

    def channel_infos(self):
        """
        .. deprecated:: 1.9.0
           Use :func:`channel_info` instead.
        """
        warn('This method is deprecated. Use "channel_info" instead.', DeprecationWarning)
        return self.channel_info()

    def channel_info(self):
        """
        Return a dictionary with the keys *NewChannel* and
        *NewPossibleChannels* indicating the active channel and
        alternative ones.
        """
        return self._action('GetChannelInfo')

    def set_channel(self, number):
        """
        Set a new channel. *number* must be a valid channel number for
        the active WLAN. (Valid numbers are listed by *alternative_channels*.)
        """
        self._action('SetChannel', NewChannel=number)

    def get_generic_host_entry(self, index):
        """
        Return a dictionary with information about the device
        internally stored at the position 'index'.
        """
        result = self._action(
            'GetGenericAssociatedDeviceInfo',
            NewAssociatedDeviceIndex=index
        )
        return result

    def get_specific_host_entry(self, mac_address):
        """
        Return a dictionary with information about the device
        with the given 'mac_address'.
        """
        result = self._action(
            'GetSpecificAssociatedDeviceInfo',
            NewAssociatedDeviceMACAddress=mac_address
        )
        return result

    def get_hosts_info(self):
        """
        Returns a list of dictionaries with information about the known
        hosts. The dict-keys are: 'service', 'index', 'status', 'mac',
        'ip', 'signal', 'speed'
        """
        information = []
        for index in itertools.count():
            try:
                host = self.get_generic_host_entry(index)
            except IndexError:
                break
            information.append({
                'service': self.service,
                'index': index,
                'status': host['NewAssociatedDeviceAuthState'],
                'mac': host['NewAssociatedDeviceMACAddress'],
                'ip': host['NewAssociatedDeviceIPAddress'],
                'signal': host['NewX_AVM-DE_SignalStrength'],
                'speed': host['NewX_AVM-DE_Speed']
            })
        return information

    def get_info(self):
        """
        Returns a dictionary with general internal information about
        the current wlan network according to the AVM documentation.
        """
        return self._action("GetInfo")

    @property
    def is_enabled(self):
        """Returns whether the guest network is enabled."""
        return self.get_info()["NewEnable"]

    def enable(self):
        """Enables the associated network."""
        self._set_enable(True)

    def disable(self):
        """Disables the associated network."""
        self._set_enable(False)

    def _set_enable(self, status):
        """Helper function for enable|disable."""
        self._action("SetEnable", arguments={"NewEnable": status})

    def get_password(self):
        """Returns the current password of the associated wlan."""
        return self._action("GetSecurityKeys")["NewKeyPassphrase"]

    def set_password(self, password=None, length=DEFAULT_PASSWORD_LENGTH):
        """
        Sets a new password for the associated wlan.
        If no password is given a new one is created with the given
        length (the new password can get read with a subsequent call of
        `get_password`). Also creates a new pre-shared key.
        """
        preshared_key = self._create_preshared_key()
        password = password or self._create_password(length)
        arguments = {
            "NewKeyPassphrase": password,
            "NewPreSharedKey": preshared_key,
            "NewWEPKey0": "",
            "NewWEPKey1": "",
            "NewWEPKey2": "",
            "NewWEPKey3": "",
        }
        self._action("SetSecurityKeys", arguments=arguments)

    def _create_preshared_key(self):
        """
        Returns a new pre-shared key for setting a new password.
        The sequence is of uppercase characters as this is default on FritzOS
        at time of writing.
        """
        info = self.get_info()
        characters = info["NewAllowedCharsPSK"]
        length = info["NewMaxCharsPSK"]
        return "".join(random.choices(characters, k=length)).upper()

    @staticmethod
    def _create_password(length):
        """
        Returns a human-readable password with the given length.
        """
        # add just two human-readable special characters.
        # password strength increases with the length.
        # character permutations are: 64**length
        characters = string.ascii_letters + string.digits + "@#"
        return "".join(random.choices(characters, k=length))


class FritzGuestWLAN(FritzWLAN):
    """
    Inherits from FritzWLAN and provides all the same methods but for
    the guest network. On devices not providing a guest network this
    class will not fail, but handle the wlan network with the highest
    internal service number (which is by default the guest network on
    guest network providing devices).

    All parameters are optional. If given, they have the following
    meaning: `fc` is an instance of FritzConnection, `address` the ip of
    the Fritz!Box, `port` the port to connect to, `user` the username,
    `password` the password, `timeout` a timeout as floating point
    number in seconds, `use_tls` a boolean indicating to use TLS
    (default False).
    """
    def __init__(self, *args, **kwargs):
        """
        Initialize the guest wlan instance. All parameters are
        optional. If given, they have the following meaning: `fc` is an
        instance of FritzConnection, `address` the ip of the Fritz!Box,
        `port` the port to connect to, `user` the username, `password`
        the password, `timeout` a timeout as floating point number in
        seconds, `use_tls` a boolean indicating to use TLS (default
        False).
        """
        super().__init__(*args, **kwargs)
        for n in itertools.count(1):
            service = self.fc.services.get(f"{SERVICE}{n}")
            if service is None:
                self.service = n - 1
                break



"""
Module to communicate with the AVM Fritz!Box.
"""
# This module is part of the FritzConnection package.
# https://github.com/kbr/fritzconnection
# License: MIT (https://opensource.org/licenses/MIT)
# Author: Klaus Bremer


import json
import os
import pickle
import string
import xml.etree.ElementTree as ElementTree
from pathlib import Path

import requests
from requests.auth import HTTPDigestAuth

from .devices import DeviceManager
from .exceptions import (
    FritzConnectionException,
    FritzResourceError,
    FritzServiceError,
)
from .soaper import Soaper
from .utils import get_bool_env

# disable InsecureRequestWarning from urllib3
# because of skipping certificate verification:
import urllib3

urllib3.disable_warnings()


# FritzConnection defaults:
FRITZ_IP_ADDRESS = "169.254.1.1"
FRITZ_TCP_PORT = 49000
FRITZ_TLS_PORT = 49443
FRITZ_USERNAME = "dslf-config"  # for Fritz!OS < 7.24
FRITZ_IGD_DESC_FILE = "igddesc.xml"
FRITZ_TR64_DESC_FILE = "tr64desc.xml"
FRITZ_DESCRIPTIONS = [FRITZ_IGD_DESC_FILE, FRITZ_TR64_DESC_FILE]
FRITZ_USERNAME_REQUIRED_VERSION = 7.24
FRITZ_APPLICATION_ACCESS_DISABLED = """\n
    FRITZ!Box: access for applications disabled.
    Check: Home Network -> Network -> Network Settings
    for "Allow access for applications".\n"""
FRITZ_CACHE_DIR = ".fritzconnection"
FRITZ_CACHE_EXT = "_cache"
FRITZ_CACHE_JSON_SUFFIX = "json"
FRITZ_CACHE_PICKLE_SUFFIX = "pcl"
FRITZ_CACHE_FORMAT_JSON = "json"
FRITZ_CACHE_FORMAT_PICKLE = "pickle"
FRITZ_CACHE_FORMATS = {
    FRITZ_CACHE_FORMAT_JSON: FRITZ_CACHE_JSON_SUFFIX,
    FRITZ_CACHE_FORMAT_PICKLE: FRITZ_CACHE_PICKLE_SUFFIX
}
FRITZ_CACHE_UNKNOWN_FORMAT_MESSAGE = f"""\
    Unknown cache format "{{}}".\n
    Use one of {tuple(FRITZ_CACHE_FORMATS.keys())}.\n"""
FRITZ_ENV_USERNAME = "FRITZ_USERNAME"
FRITZ_ENV_PASSWORD = "FRITZ_PASSWORD"
FRITZ_ENV_USECACHE = "FRITZ_USECACHE"
FRITZ_ENV_CACHE_FORMAT = "FRITZ_CACHEFORMAT"
FRITZ_ENV_CACHEDIRECTORY = "FRITZ_CACHEDIRECTORY"

# same defaults as used by requests:
DEFAULT_POOL_CONNECTIONS = 10
DEFAULT_POOL_MAXSIZE = 10

# supported protocols:
PROTOCOLS = ['http://', 'https://']


class FritzConnection:
    """
    Main class to set up a connection to the Fritz!Box router. All
    parameters are optional. `address` should be the ip of a router, in
    case that are multiple Fritz!Box routers in a network, the ip must
    be given. Otherwise, it is undefined which router will respond. If
    `user` and `password` are not provided, the environment gets checked for
    FRITZ_USERNAME and FRITZ_PASSWORD settings and taken from there, if
    found.

    The optional parameter `timeout` is a floating number in seconds
    limiting the time waiting for a router response. This is a global
    setting for the internal communication with the router. In case of a
    timeout a `requests.ConnectTimeout` exception gets raised.

    .. versionadded:: 1.1

    `use_tls` accepts a boolean for using encrypted communication with
    the Fritz!Box. Default is `False`.

    .. versionadded:: 1.2

    For some actions the Fritz!Box needs a password and since Fritz!OS
    7.24 also requires a username, the previous default username is just
    valid for OS versions < 7.24. In case the username is not given and
    the system version is 7.24 or newer, FritzConnection uses the last
    logged-in username as default.

    .. versionadded:: 1.5

    For applications where the urllib3 default connection-pool size
    should get adapted, the arguments `pool_connections` and
    `pool_maxsize` can get set explicitly.

    .. versionadded:: 1.6

    The flag `use_cache` activates caching (default `False`). Caching
    can speed up instanciation significantly. The cached data are
    specific for the router ip, the router model and the installed
    FritzOS version. Multiple devices in the network can have separate
    cache-fies and can get used in parallel. By default the cache files
    are stored in the user home-directory in a `.fritzconnection`
    folder. To change this location use the parameter `cache_directory`
    providing a string or a `pathlib.Path` object. With `cache_format`
    two formats can specified for data serialization: `json` and
    `pickle`. These two values are available as constants
    `FRITZ_CACHE_FORMAT_JSON` and `FRITZ_CACHE_FORMAT_PICKLE`. Default
    is `pickle`. The flag `verify_cache` will enable cache verification
    (default is `True`). If set to `False` loading the api-data will be
    even faster, but the cache will not get renewed in case of FritzOS
    updates or a router change. All cache-settings can also configured
    in the environment: FRITZ_USECACHE (True|False), FRITZ_CACHEFORMAT
    (json|pickle) and FRITZ_CACHEDIRECTORY (a path).

    .. versionadded:: 1.10
    """

    def __init__(
        self,
        address=None,
        port=None,
        user=None,
        password=None,
        timeout=None,
        use_tls=False,
        use_cache=False,
        verify_cache=True,
        cache_directory=None,
        cache_format=None,
        pool_connections=DEFAULT_POOL_CONNECTIONS,
        pool_maxsize=DEFAULT_POOL_MAXSIZE,
    ):
        """
        Initialisation of FritzConnection: reads all data from the box
        and also the api-description (the servicenames and according
        actionnames as well as the parameter-types) that can vary among
        models and stores the information as instance-attributes. This
        can be an expensive operation. Because of this an instance of
        FritzConnection should be created once and reused in an
        application.

        All parameters are optional. But if there is more than one
        FritzBox in the network, an address (ip as string) must be
        given, otherwise it is not defined which box may respond. If no
        user is given the Environment gets checked for a FRITZ_USERNAME
        setting. If there is no entry in the environment the
        avm-default-username will be used. If no password is given the
        Environment gets checked for a FRITZ_PASSWORD setting. So
        password can be used without using configuration-files or even
        hardcoding.

        The optional parameter `timeout` is a floating point number in
        seconds limiting the time waiting for a router response. The
        timeout can also be a tuple for different values for connection-
        and read-timeout values: (connect timeout, read timeout). The
        timeout is a global setting for the internal communication with
        the router. In case of a timeout a `requests.ConnectTimeout`
        exception gets raised.

        `use_tls` accepts a boolean for using encrypted communication
        with the Fritz!Box. Default is `False`.

        `use_cache` is a boolean whether a cache should get used for the
        router api data. By default the api data are loaded from the
        router at instanciation time what can take several seconds to
        complete. `cache_directory` is the path to the directory storing
        the cached data. By default this is a folder named
        '.fritzconnection' in the users home-directory. `cache_format`
        supports two file-formats: json and pickle (default). All cache
        settings can also defined in the environment: FRITZ_USECACHE
        (True|False), FRITZ_CACHEFORMAT (json|pickle) and
        FRITZ_CACHEDIRECTORY (a path).

        If `verify_cache` is `True` it checks whether the model has
        changed or the system software has got an update. In this case
        the cache gets renewed. Default is `True`. Deactivating the
        verification gives another gain in speed, but on updates and
        other changes you are on your own.

        `pool_connections` and `pool_maxsize` accept integers for
        changing the default urllib3 settings in order to modify the
        number of reusable connections.
        """
        if address is None:
            address = FRITZ_IP_ADDRESS
        if user is None:
            user = os.getenv(FRITZ_ENV_USERNAME, FRITZ_USERNAME)
        if password is None:
            password = os.getenv(FRITZ_ENV_PASSWORD, "")
        if use_cache is None:
            use_cache = get_bool_env(FRITZ_ENV_USECACHE)
        if cache_format is None:
            cache_format = os.getenv(
                FRITZ_ENV_CACHE_FORMAT, FRITZ_CACHE_FORMAT_PICKLE
            )
        if cache_directory is None:
            cache_directory = os.getenv(FRITZ_ENV_CACHEDIRECTORY, None)
        if port is None and use_tls:
            port = FRITZ_TLS_PORT
        elif port is None:
            port = FRITZ_TCP_PORT
        address = self.set_protocol(address, use_tls)

        # a session will speed up connections (significantly for tls)
        # and is required to change the default poolsize:
        session = requests.Session()
        session.verify = False
        if password:
            session.auth = HTTPDigestAuth(user, password)
        adapter = requests.adapters.HTTPAdapter(
            pool_connections=pool_connections, pool_maxsize=pool_maxsize)
        session.mount(PROTOCOLS[use_tls], adapter)
        # store as instance attributes for use by library modules
        self.address = address
        self.session = session
        self.timeout = timeout
        self.port = port

        self.soaper = Soaper(
            address, port, user, password, timeout=timeout, session=session
        )
        self.device_manager = DeviceManager(timeout=timeout, session=session)
        self._load_router_api(
            use_cache, cache_directory, cache_format, verify_cache
        )
        # set default user for FritzOS >= 7.24:
        self._reset_user(user, password)

    def __repr__(self):
        """Return a readable representation"""
        return (
            f"{self.modelname} at {self.soaper.address}\n"
            f"FRITZ!OS: {self.system_version}"
        )

    @property
    def services(self):
        """
        Dictionary of service instances. Keys are the service names.
        """
        return self.device_manager.services

    @property
    def modelname(self):
        """
        Returns the modelname of the router.
        """
        return self.device_manager.modelname

    @property
    def system_version(self):
        """
        Returns system version if known, otherwise None.
        """
        return self.device_manager.system_version

    @property
    def device_description(self):
        """
        Returns a string with the device description. This is a
        combination of the device model name and the installed software
        version.
        """
        return self.call_action("DeviceInfo1", "GetInfo")["NewDescription"]

    @staticmethod
    def normalize_name(name):
        """
        Returns the normalized service name. E.g. `WLANConfiguration` and
        `WLANConfiguration:1` will get converted to `WLANConfiguration1`.
        """
        if ":" in name:
            name, number = name.split(":", 1)
            name = name + number
        elif name[-1] not in string.digits:
            name = name + "1"
        return name

    @staticmethod
    def set_protocol(url, use_tls):
        """
        Sets the protocol of the `url` according to the `use_tls`-flag
        and returns the modified `url`. Does not check whether the `url`
        given as parameter is correct.
        """
        url = url.split("//", 1)[-1]
        return PROTOCOLS[use_tls] + url

    def _reset_user(self, user, password):
        """
        For Fritz!OS >= 7.24: if a password is given and the username is
        the historic FRITZ_USERNAME, then check for the last logged-in
        username and use this username for the soaper. Also recreate the
        session used by the soaper and the device_manager.

        This may not guarantee a valid user/password combination, but is
        the way AVM recommends setting the required username in case a
        username is not provided.
        """
        try:
            sys_version = float(self.system_version)
        except (ValueError, TypeError):
            # version not available: don't do anything
            return
        if (sys_version >= FRITZ_USERNAME_REQUIRED_VERSION
            and user == FRITZ_USERNAME
            and password
        ):
            last_user = None
            response = self.call_action(
                'LANConfigSecurity1', 'X_AVM-DE_GetUserList'
            )
            root = ElementTree.fromstring(response['NewX_AVM-DE_UserList'])
            for node in root:
                if node.tag == 'Username' and node.attrib['last_user'] == '1':
                    last_user = node.text
                    break
            if last_user is not None:
                self.session.auth = HTTPDigestAuth(last_user, password)
                self.soaper.user = last_user
                self.soaper.session = self.session
                self.device_manager.session = self.session

    # -------------------------------------------
    # public api:
    # -------------------------------------------

    def call_action(
        self,
        service_name,
        action_name,
        *,
        arguments=None,
        **kwargs
    ):
        """
        Executes the given action of the given service. Both parameters
        are required. Arguments are optional and can be provided as a
        dictionary given to 'arguments' or as separate keyword
        parameters. If 'arguments' is given additional
        keyword-parameters as further arguments are ignored.

        The argument values can be of type *str*, *int* or *bool*.
        (Note: *bool* is provided since 1.3. In former versions booleans
        must be provided as numeric values: 1, 0).

        If the service_name does not end with a number-character (like
        "1""), a "1" gets added by default. If the service_name ends
        with a colon and a number, the colon gets removed. So i.e.
        "WLANConfiguration" expands to "WLANConfiguration1" and
        "WLANConfiguration:2" converts to "WLANConfiguration2"". Invalid
        service names will raise a ServiceError and invalid action names
        will raise an ActionError.
        """
        arguments = arguments if arguments else dict()
        if not arguments:
            arguments.update(kwargs)
        service_name = self.normalize_name(service_name)
        try:
            service = self.device_manager.services[service_name]
        except KeyError:
            raise FritzServiceError(f'unknown service: "{service_name}"')
        return self.soaper.execute(service, action_name, arguments)

    def reconnect(self):
        """
        Terminate the connection and reconnects with a new ip.
        """
        self.call_action("WANIPConn1", "ForceTermination")

    def reboot(self):
        """
        Reboot the system.
        """
        self.call_action("DeviceConfig1", "Reboot")

    # -------------------------------------------
    # internal methods to load router-api:
    # -------------------------------------------

    def _load_router_api(
        self,
        use_cache=False,
        cache_directory=None,
        cache_format=FRITZ_CACHE_FORMAT_JSON,
        verify_cache=True,
    ):
        """
        Load the router api.

        If `use_cache` is `False`, load the api from the router. If
        `use_cache` is `True``, the api data are loaded from a file,
        which is in pickle or json format, according to the setting of
        `cache_format`.  The file location can get set by the argument
        `cache_directory`. If `cache_directory` is not given, the
        default-directory is used (which is in most cases a subdirectory
        of the user home directory). After loading from a file the
        cached data are checked to detect a software-update or a change
        of the router model. If this check fails, the api gets reloaded
        from the router and the cache data are updated. The same happens
        on errors loading the cache-file.
        """
        def reload_api():
            # reset in case of remaining artefacts:
            self.device_manager.descriptions = []
            self.device_manager.services = {}
            # reload and save again:
            self._load_api_from_router()
            self._write_api_to_cache(path, cache_format)

        if use_cache:
            path = self._get_cache_path(cache_directory, cache_format)
            try:
                self._load_api_from_cache(path, cache_format)
            except FileNotFoundError:
                # can happen i.e. on first run
                reload_api()
            else:
                if verify_cache and not self._is_valid_cache():
                    # can happen on model changes or software updates
                    reload_api()
        else:
            self._load_api_from_router()

    def _is_valid_cache(self):
        """
        Checks whether the cache-data seems to be valid. Returns a
        booean: `True` if valid, `False` otherwise.
        """
        # system_id is something like ('FRITZ!Box 7590', '154.07.29') which
        # originates from the device description and is part of the cache data.
        cached_id = (
            self.device_manager.modelname,
            self.device_manager.system_info[-1]
        )
        # retrive the same information from an api call. If the result
        # is the same, the cache-data can considered as valid.
        try:
            device_info = self.call_action("DeviceInfo:1", "GetInfo")
        except FritzConnectionException:
            # something went wrong, cache data can not be verified.
            return False
        current_id = (
            device_info['NewModelName'],
            device_info['NewSoftwareVersion']
        )
        return cached_id == current_id

    def _get_cache_path(self, cache_directory, cache_format):
        """
        Returns the path to the cache file (including the filename and
        extension) as a Path instance.
        """
        # ignore optional scheme:
        address = self.address.split('//')[-1]
        address = address.replace(".", "_")
        try:
            suffix = FRITZ_CACHE_FORMATS[cache_format]
        except KeyError:
            message = FRITZ_CACHE_UNKNOWN_FORMAT_MESSAGE.format(cache_format)
            raise FritzConnectionException(message)
        filename = f"{address}{FRITZ_CACHE_EXT}.{suffix}"
        if cache_directory:
            return Path(cache_directory) / filename
        return Path().home() / FRITZ_CACHE_DIR /filename

    def _write_api_to_cache(self, path, cache_format):
        """
        Stores the api data in a cache-file.
        """
        binary = "wb"
        text = "wt"
        mode = binary if cache_format == FRITZ_CACHE_FORMAT_PICKLE else text
        with open(path, mode) as fobj:
            if mode == binary:
                pickle.dump(self.device_manager.descriptions, fobj)
            else:
                json.dump(self.device_manager.serialize(), fobj)

    def _load_api_from_cache(self, path, cache_format):
        """
        Read the api data from a cache-file and forwards the data to the
        device_manager.
        Currently two formats are supported: `pickle` and `json`. If
        `cache_format` is not `pickle` it is assumed that the format is
        `json`.
        Raise a FileNotFoundError in case of an invalide path.
        """
        binary = "rb"
        text = "rt"
        mode = binary if cache_format == FRITZ_CACHE_FORMAT_PICKLE else text
        with open(path, mode) as fobj:
            if mode == binary:
                self.device_manager.descriptions = pickle.load(fobj)
            else:
                self.device_manager.deserialize(json.load(fobj))
        self.device_manager.scan()

    def _load_api_from_router(self):
        """
        Read the api data from the router and forwards the data to the
        device_manager.
        """
        for description in FRITZ_DESCRIPTIONS:
            source = f"{self.address}:{self.port}/{description}"
            try:
                self.device_manager.add_description(source)
            except FritzResourceError:
                # resource not available:
                # this can happen on devices not providing
                # an igddesc-file.
                # ignore this
                # But if the "tr64desc.xml" file is missing the router
                # may not have TR-064 activated. In this case raise a
                # useful error-message.
                if description == FRITZ_TR64_DESC_FILE:
                    raise FritzConnectionException(
                        FRITZ_APPLICATION_ACCESS_DISABLED
                    )
        self.device_manager.scan()
        self.device_manager.load_service_descriptions(self.address, self.port)



"""
Module to communicate with the AVM Fritz!Box service providing real time
phone-call events.

To run fritzmonitor, the CallMonitor service of the box has to be activated.
This can be done with any registered Phone by typing the following codes:
activate: #96*5*
deactivate: #96*4* 
"""
# This module is part of the FritzConnection package.
# https://github.com/kbr/fritzconnection
# License: MIT (https://opensource.org/licenses/MIT)
# Author: Klaus Bremer


import queue
import socket
import threading
import time


FRITZ_IP_ADDRESS = "169.254.1.1"
FRITZ_MONITOR_PORT = 1012
FRITZ_MONITOR_QUEUE_SIZE = 256
FRITZ_MONITOR_CHUNK_SIZE = 1024 * 4
FRITZ_MONITOR_SOCKET_TIMEOUT = 10

MIN_RECONNECT_DELAY = 0.02  # minimum delay in seconds
MAX_RECONNECT_DELAY = 60  # maximum delay in seconds
RECONNECT_DELAY_FACTOR = 10  # factor to increase delays between reconnection tries
RECONNECT_TRIES = 10  # number of tries to reconnect before giving up


def delayer(
    min_delay=MIN_RECONNECT_DELAY,
    max_delay=MAX_RECONNECT_DELAY,
    multiplier=RECONNECT_DELAY_FACTOR,
):
    """
    delay generator with increasing sleep-times.
    """
    delay = min(min_delay, max_delay)
    while True:
        time.sleep(delay)
        yield
        delay = min(delay * multiplier, max_delay)


class EventReporter:
    """
    Takes a Queue and implements a buffer for line-separated data.
    If at least one line is in the buffer, the line gets put into the 
    Queue for further processing elsewhere (by a routine reading the queue).
    """

    def __init__(self, monitor_queue, block_on_filled_queue=False):
        """
        Takes the monitor queue (of type queue.Queue) and a boolean
        'block_on_filled_queue'. If 'block_on_filled_queue' is True the thread
        will wait until a free slot is available blocking the thread reading
        call_monitor events from the socket connected to the Fritz!Box.
        """
        self.buffer = ""
        self.monitor_queue = monitor_queue
        self.block_on_filled_queue = block_on_filled_queue

    def add(self, data):
        """
        Adds the given 'data' to the buffer. If the buffer holds at least one
        line (separated by newline-character), the line (or lines) are put into the
        'monitor_queue' as events to be processed from a reader elsewhere.
        """
        self.buffer += data
        *parts, self.buffer = self.buffer.split("\n")
        for part in parts:
            try:
                self.monitor_queue.put(part, block=self.block_on_filled_queue)
            except queue.Full:
                # ignore
                pass


class FritzMonitor:
    """
    Monitor Fritz!Box events about started, finished, incoming and outgoing
    calls.
    """

    def __init__(
        self,
        address=FRITZ_IP_ADDRESS,
        port=FRITZ_MONITOR_PORT,
        timeout=FRITZ_MONITOR_SOCKET_TIMEOUT,
        encoding="utf-8",
    ):
        self.address = address
        self.port = port
        self.timeout = timeout
        self.stop_flag = threading.Event()
        self.monitor_thread = None
        self.encoding = encoding

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.stop()

    @property
    def has_monitor_thread(self):
        """
        Returns True if a monitor-thread has been created.
        That should be the case after calling start() and before calling stop().
        """
        return bool(self.monitor_thread)

    @property
    def is_alive(self):
        """
        Returns True if there is a monitor-thread and the thread is running.
        Returns False otherwise.
        """
        return self.has_monitor_thread and self.monitor_thread.is_alive()

    def start(
        self,
        queue_size=FRITZ_MONITOR_QUEUE_SIZE,
        block_on_filled_queue=False,
        reconnect_delay=MAX_RECONNECT_DELAY,
        reconnect_tries=RECONNECT_TRIES,
        sock=None,
    ):
        """
        Start the monitor thread and return a Queue instance with the given size
        to report the call_monitor events. Events are of type string. Raises an
        `OSError` if the socket can not get connected in a given timeout. Raises
        a `RuntimeError` if start() get called a second time without calling
        stop() first. `queue_size` is the number of events the queue can store.
        If `block_on_filled_queue` is False the event will get discarded in case
        of no free block (default). On True the EventReporter will block until a
        slot is available. `reconnect_delay` defines the maximum time interval in
        seconds between reconnection tries, in case that a socket-connection
        gets lost. `reconnect_tries` defines the number of consecutive to
        reconnect a socket before giving up.
        """
        if self.monitor_thread:
            # It's an error to create a second thread for monitoring
            raise RuntimeError("A FritzMonitor thread is already running")
        # get socket or raise OSError in main thread:
        sock = self._get_connected_socket(sock=sock)
        monitor_queue = queue.Queue(maxsize=queue_size)
        kwargs = {
            "monitor_queue": monitor_queue,
            "sock": sock,
            "block_on_filled_queue": block_on_filled_queue,
            "reconnect_delay": reconnect_delay,
            "reconnect_tries": reconnect_tries,
        }
        # clear event object in case the instance gets 'reused':
        self.stop_flag.clear()
        self.monitor_thread = threading.Thread(target=self._monitor, kwargs=kwargs)
        self.monitor_thread.start()
        return monitor_queue

    def stop(self):
        """
        Stop the current running monitor_thread.
        """
        if self.monitor_thread:
            if self.monitor_thread.is_alive():
                self.stop_flag.set()  # tell thread to terminate
                self.monitor_thread.join()  # wait for termination without timeout
            self.monitor_thread = None

    def _get_connected_socket(self, sock=None):
        """
        Takes the given socket and builds a connection to address and port defined
        at instantiation. If no socket is given a new one gets created.
        Returns the socket.
        """
        if sock is None:
            sock = socket.socket()
            # these options should work on all platforms:
            sock.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)
            # socket actions can fail for different reasons,
            # without a timeout they can hang a long time.
            sock.settimeout(self.timeout)
        try:
            sock.connect((self.address, self.port))
        except socket.timeout:
            # will fail if the address:port is not available (not present or in use)
            # socket.timeout is subclass of OSError, so it is save to raise an OSError
            msg = f"Unable to connect to '{self.address}:{self.port}'"
            raise OSError(msg)
        return sock

    def _reconnect_socket(
        self,
        sock,
        max_reconnect_delay=MAX_RECONNECT_DELAY,
        reconnect_tries=RECONNECT_TRIES,
    ):
        """
        Try to reconnect a lost connection on the given socket.
        Returns True on success and False otherwise.
        """
        reconnect_delay = delayer(max_delay=max_reconnect_delay)
        while reconnect_tries > 0:
            next(reconnect_delay)
            try:
                self._get_connected_socket(sock)
            except OSError:
                reconnect_tries -= 1
            else:
                return True
        return False

    def _monitor(
        self,
        monitor_queue,
        sock,
        block_on_filled_queue,
        reconnect_delay,
        reconnect_tries,
    ):
        """
        The internal monitor routine running in a separate thread.
        """
        # Instantiate an EventReporter to push event to the event_queue.
        event_reporter = EventReporter(
            monitor_queue=monitor_queue, block_on_filled_queue=block_on_filled_queue
        )
        while not self.stop_flag.is_set():
            try:
                raw_data = sock.recv(FRITZ_MONITOR_CHUNK_SIZE)
            except socket.timeout:
                # without a timeout an open socket will never return from a
                # connection closed by a router (may be of limited resources).
                # Therefore be sure to set a timeout at socket creation (elsewhere).
                # So just try again after timeout.
                continue
            if not raw_data:
                # empty response indicates a lost connection.
                # try to reconnect.
                success = self._reconnect_socket(
                    sock,
                    max_reconnect_delay=reconnect_delay,
                    reconnect_tries=reconnect_tries,
                )
                if not success:
                    # reconnect has failed: terminate the thread
                    break
            else:
                # sock.recv returns a bytearray to decode:
                response = raw_data.decode(self.encoding)
                event_reporter.add(response)
        # clean up on terminating thread:
        try:
            sock.close()
        except OSError:
            pass
        # reset monitor_thread to be able to restart
        self.monitor_thread = None



""""
Module to inspect phone-call events in real time.

This is the command line interface for the core.fritzmonitor module
and should serve as an example how to use an instance of FritzMonitor.

To run this, the CallMonitor service of the box has to be activated.
This can be done with any registered Phone by typing the following codes:
activate: #96*5*
deactivate: #96*4* 
"""
# This module is part of the FritzConnection package.
# https://github.com/kbr/fritzconnection
# License: MIT (https://opensource.org/licenses/MIT)
# Author: Klaus Bremer


import argparse
import queue

from ..core.fritzmonitor import (
    FritzMonitor,
    FRITZ_IP_ADDRESS,
    FRITZ_MONITOR_SOCKET_TIMEOUT,
)
from .. import __version__


HEALTHCHECK_TIMEOUT = 10


def print_header(args):
    print(f"\nfritzconnection v{__version__}")
    print(f"start fritzmonitor on address: {args.address}")
    print(f"settings for socket-timeout: {args.timeout} [sec]")
    print(f"settings for healthcheck-timeout: {args.healthcheck} [sec]")
    print("(to stop press ^C)\n")


def get_cli_arguments():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-i",
        "--ip-address",
        nargs="?",
        default=FRITZ_IP_ADDRESS,
        const=None,
        dest="address",
        help="Specify ip-address of the FritzBox to connect to."
        "Default: %s" % FRITZ_IP_ADDRESS,
    )
    parser.add_argument(
        "-t",
        "--timeout",
        nargs="?",
        type=int,
        default=FRITZ_MONITOR_SOCKET_TIMEOUT,
        const=None,
        dest="timeout",
        help="Setting for socket timeout [sec]."
        "Default: %s" % FRITZ_MONITOR_SOCKET_TIMEOUT,
    )
    parser.add_argument(
        "-c",
        "--healthcheck",
        nargs="?",
        type=int,
        default=HEALTHCHECK_TIMEOUT,
        const=None,
        dest="healthcheck",
        help="Setting for internal health-check interval [sec]."
        "Default: %s" % HEALTHCHECK_TIMEOUT,
    )
    args = parser.parse_args()
    return args


def process_events(monitor, event_queue, healthcheck_interval):
    while True:
        try:
            event = event_queue.get(timeout=healthcheck_interval)
        except queue.Empty:
            # check health:
            if not monitor.is_alive:
                raise OSError("Error: fritzmonitor connection failed")
        else:
            # do event processing here:
            print(event)


def main():
    """
    Entry point: example to use FritzMonitor.
    """
    args = get_cli_arguments()
    print_header(args)
    # create a FritzMonitor instance and get the event_queue by calling start().
    # start() returns the queue for the events.
    try:
        with FritzMonitor(address=args.address, timeout=args.timeout) as monitor:
            event_queue = monitor.start()
            process_events(monitor, event_queue, healthcheck_interval=args.healthcheck)
    except (OSError, KeyboardInterrupt) as err:
        print(err)
    print("exit fritzmonitor")


if __name__ == "__main__":
    main()



"""
Abstract base class for all library classes.

An internal module providing the `AbstractLibraryBase` class. This is an abstract class that should not get instantiated but should serve as a base class for library classes providing a common initialisation.

"""
# This module is part of the FritzConnection package.
# https://github.com/kbr/fritzconnection
# License: MIT (https://opensource.org/licenses/MIT)
# Author: Klaus Bremer


from ..core.fritzconnection import (
    FritzConnection,
    DEFAULT_POOL_CONNECTIONS,
    DEFAULT_POOL_MAXSIZE,
)


class AbstractLibraryBase:
    """
    Abstract base class for library classes. Implements the common
    initialisation. All parameters are optional. If given, they have the
    following meaning: `fc` is an instance of FritzConnection, `address`
    the ip of the Fritz!Box, `port` the port to connect to, `user` the
    username, `password` the password, `timeout` a timeout as floating
    point number in seconds, `use_tls` a boolean indicating to use TLS
    (default False). `pool_connections` and `pool_maxsize` are to
    override the urllib3 default settings.
    """
    def __init__(self, fc=None, address=None, port=None, user=None,
                 password=None, timeout=None, use_tls=False,
                 pool_connections=DEFAULT_POOL_CONNECTIONS,
                 pool_maxsize=DEFAULT_POOL_MAXSIZE,
        ):
        if fc is None:
            kwargs = {
                k:v for k, v in locals().items() if k not in ('self', 'fc')
            }
            fc = FritzConnection(**kwargs)
        self.fc = fc

    @property
    def modelname(self):
        """
        The device modelname. Every library module derived from
        `AbstractLibraryBase` inherits this property.
        """
        return self.fc.modelname



"""
Module to access and control the known hosts.
"""
# This module is part of the FritzConnection package.
# https://github.com/kbr/fritzconnection
# License: MIT (https://opensource.org/licenses/MIT)
# Author: Klaus Bremer


import itertools

from ..core.exceptions import (
    FritzActionError,
    FritzArgumentError,
    FritzLookUpError,
)
from ..core.processor import HostStorage
from ..core.utils import get_xml_root
from .fritzbase import AbstractLibraryBase


SERVICE = "Hosts1"


class FritzHosts(AbstractLibraryBase):
    """
    Class to access the registered hosts. All parameters are optional. If
    given, they have the following meaning: `fc` is an instance of
    FritzConnection, `address` the ip of the Fritz!Box, `port` the port
    to connect to, `user` the username, `password` the password,
    `timeout` a timeout as floating point number in seconds, `use_tls` a
    boolean indicating to use TLS (default False).
    """

    def _action(self, actionname, *, arguments=None, **kwargs):
        return self.fc.call_action(SERVICE, actionname, arguments=arguments, **kwargs)

    @property
    def host_numbers(self):
        """The number of known hosts."""
        result = self._action("GetHostNumberOfEntries")
        return result["NewHostNumberOfEntries"]

    def get_generic_host_entry(self, index):
        """
        Returns a dictionary with information about a device internally
        registered by the position *index*. Index-positions are
        zero-based.
        """
        return self._action("GetGenericHostEntry", NewIndex=index)

    def get_generic_host_entries(self):
        """
        Generator returning a dictionary for every host as provided by
        `get_generic_host_entry()`. (See also `get_hosts_info()` that
        returns a list of dictionaries with different key-names.)
        """
        for index in itertools.count():
            try:
                yield self.get_generic_host_entry(index)
            except IndexError:
                break

    def get_specific_host_entry(self, mac_address):
        """
        Returns a dictionary with information about a device addressed
        by the MAC-address.
        """
        return self._action("GetSpecificHostEntry", NewMACAddress=mac_address)

    def get_specific_host_entry_by_ip(self, ip):
        """
        Returns a dictionary with information about a device addressed
        by the ip-address. Provides additional information about
        connection speed and system-updates for AVM devices.
        """
        return self._action("X_AVM-DE_GetSpecificHostEntryByIP", NewIPAddress=ip)

    def get_host_status(self, mac_address):
        """
        Provides status information about the device with the given
        `mac_address`. Returns `True` if the device is active or `False`
        otherwise. Returns `None` if the device is not known or the
        `mac_address` is invalid.
        """
        try:
            result = self.get_specific_host_entry(mac_address)
        except (FritzArgumentError, FritzLookUpError):
            return None
        return result["NewActive"]

    def get_active_hosts(self):
        """
        Returns a list of dicts with information about the active
        devices. The dict-keys are: 'ip', 'name', 'mac', 'status',
        'interface_type', 'address_source', 'lease_time_remaining'
        """
        return [host for host in self.get_hosts_info() if host["status"]]

    def get_hosts_info(self):
        """
        Returns a list of dicts with information about the known hosts.
        The dict-keys are: 'ip', 'name', 'mac', 'status',
        'interface_type', 'address_source', 'lease_time_remaining'.
        """
        result = []
        for index in itertools.count():
            try:
                host = self.get_generic_host_entry(index)
            except IndexError:
                # no more host entries:
                break
            result.append(
                {
                    "ip": host["NewIPAddress"],
                    "name": host["NewHostName"],
                    "mac": host["NewMACAddress"],
                    "status": host["NewActive"],
                    "interface_type": host["NewInterfaceType"],
                    "address_source": host["NewAddressSource"],
                    "lease_time_remaining": host["NewLeaseTimeRemaining"],
                }
            )
        return result

    def get_mesh_topology(self, raw=False):
        """
        Returns information about the mesh network topology. If `raw` is
        `False` the topology gets returned as a dictionary with a list
        of nodes. If `raw` is `True` the data are returned as text in
        json format. Default is `False`.
        """
        result = self._action("X_AVM-DE_GetMeshListPath")
        path = result["NewX_AVM-DE_MeshListPath"]
        url = f"{self.fc.address}:{self.fc.port}{path}"
        with self.fc.session.get(url) as response:
            if not response.ok:
                message = f"Error {response.status_code}: Device has no access to topology information."
                raise FritzActionError(message)
            return response.text if raw else response.json()

    def get_wakeonlan_status(self, mac_address):
        """
        Returns a boolean whether wake on LAN signal gets send to the
        device with the given `mac_address` in case of a remote access.
        """
        info = self._action(
            "X_AVM-DE_GetAutoWakeOnLANByMACAddress", NewMACAddress=mac_address
        )
        return info["NewAutoWOLEnabled"]

    def set_wakeonlan_status(self, mac_address, status=False):
        """
        Sets whether a wake on LAN signal should get send send to the
        device with the given `mac_address` in case of a remote access.
        `status` is a boolean, default value is `False`. This method has
        no return value.
        """
        args = {
            "NewMACAddress": mac_address,
            "NewAutoWOLEnabled": status,
        }
        self._action("X_AVM-DE_SetAutoWakeOnLANByMACAddress", arguments=args)

    def set_host_name(self, mac_address, name):
        """
        Sets the hostname of the device with the given `mac_address` to
        the new `name`.
        """
        args = {
            "NewMACAddress": mac_address,
            "NewHostName": name,
        }
        self._action("X_AVM-DE_SetHostNameByMACAddress", arguments=args)

    def get_host_name(self, mac_address):
        """
        Returns a String with the host_name of the device with the given
        mac_address
        """
        return self.get_specific_host_entry(mac_address)["NewHostName"]

    def run_host_update(self, mac_address):
        """
        Triggers the host with the given `mac_address` to run a system
        update. The method returns immediately, but for the device it
        takes some time to do the OS update. All vendor warnings about running a
        system update apply, like not turning power off during a system
        update. So run this command with caution.
        """
        self._action("X_AVM-DE_HostDoUpdate", NewMACAddress=mac_address)

    def get_hosts_attributes(self):
        """
        Returns a list of dictionaries with information about all hosts.

        This differs from `get_hosts_info` as the information origins
        from a different FritzOS call: the information about the current
        host stats is provided by a Lua script returning an xml-stream.
        `get_hosts_attributes` triggers the Lua script and converts the
        returned data in a list of dictionaries describing the known
        hosts.

        .. versionadded:: 1.10
        """
        result = self._action("X_AVM-DE_GetHostListPath")
        url = result["NewX_AVM-DE_HostListPath"]
        storage = HostStorage(get_xml_root(url, self.fc.session))
        return storage.hosts_attributes



"""
fritzcall.py

Module to inspect the FritzBox phone API.
CLI interface.

This module is part of the FritzConnection package.
https://github.com/kbr/fritzconnection
License: MIT (https://opensource.org/licenses/MIT)
Author: Klaus Bremer
"""

from ..lib.fritzcall import FritzCall
from . utils import get_cli_arguments, get_instance, print_header


def report_calls(fc, arguments):
    print('FritzCall:')
    days = arguments.days
    num = arguments.num if not days else None
    if arguments.type == 'in':
        calls = fc.get_received_calls(num=num, days=days)
    elif arguments.type == 'out':
        calls = fc.get_out_calls(num=num, days=days)
    elif arguments.type == 'missed':
        calls = fc.get_missed_calls(num=num, days=days)
    else:
        calls = fc.get_calls(num=num, days=days)
    call_type = arguments.type if arguments.type else 'all'
    print('List of calls:', call_type, '\n')
    type_ = 'type'
    number = 'number'
    time = 'date/time'
    duration = 'duration'
    print(f'{type_:>6}   {number:24}{time:>18}{duration:>12}\n')
    for call in calls:
        print(call)
    print()


def dial_number(fc, number):
    print(f'dialing number: {number}')
    fc.dial(number)
    print('dialing done, please wait for signal.')


def add_arguments(parser):
    parser.add_argument('-n', '--num',
                        nargs='?', default=None, const=None,
                        help='max number of calls in the call-list')
    parser.add_argument('-d', '--days',
                        nargs='?', default=None, const=None,
                        help='number of days to look back for calls.')
    parser.add_argument('-t', '--type',
                        nargs='?', default=None, const=None,
                        help='type of calls: [in|out|missed]')
    parser.add_argument('-c', '--call',
                        nargs='?', default=None, const=None,
                        help='phone number to call')


def main():
    arguments = get_cli_arguments(add_arguments)
    if not arguments.password:
        print('Exit: password required.')
        return
    fc = get_instance(FritzCall, arguments)
    if arguments.call:
        dial_number(fc, arguments.call)
    else:
        print_header(fc)
        report_calls(fc, arguments)


if __name__ == '__main__':
    main()



"""
Module to access home-automation devices
"""
# This module is part of the FritzConnection package.
# https://github.com/kbr/fritzconnection
# License: MIT (https://opensource.org/licenses/MIT)
# Author: Klaus Bremer


import itertools
from warnings import warn

from .fritzbase import AbstractLibraryBase


SERVICE = 'X_AVM-DE_Homeauto1'


class FritzHomeAutomation(AbstractLibraryBase):
    """
    Interface for fritzbox homeauto service. All parameters are
    optional. If given, they have the following meaning: `fc` is an
    instance of FritzConnection, `address` the ip of the Fritz!Box,
    `port` the port to connect to, `user` the username, `password` the
    password, `timeout` a timeout as floating point number in seconds,
    `use_tls` a boolean indicating to use TLS (default False).
    """

    def _action(self, actionname, *, arguments=None, **kwargs):
        if arguments is None:
            arguments = kwargs
        return self.fc.call_action(SERVICE, actionname, arguments=arguments)

    @property
    def get_info(self):
        """
        Return a dictionary with a single key-value pair:
        'NewAllowedCharsAIN': string with all allowed chars for state
        variable AIN
        """
        return self._action('GetInfo')

    def get_device_information_by_index(self, index):
        """
        Return a dictionary with all device arguments according to the
        AVM documentation (x_homeauto) at the given internal index.
        Raise a FritzArrayIndexError (subclass of IndexError) on invalid
        index values.
        """
        return self._action('GetGenericDeviceInfos', NewIndex=index)

    def get_device_information_by_identifier(self, identifier):
        """
        Returns a dictionary with all device arguments according to the
        AVM documentation (x_homeauto) with the given identifier (AIN).
        Raise an FritzArgumentError on invalid identifier.
        """
        return self._action('GetSpecificDeviceInfos', NewAIN=identifier)

    def device_informations(self):
        """
        .. deprecated:: 1.9.0
           Use :func:`device_information` instead.
        """
        warn('This method is deprecated. Use "device_information" instead.', DeprecationWarning)
        return self.device_information()

    def device_information(self):
        """
        Returns a list of dictionaries for all known homeauto-devices.
        """
        info = list()
        for n in itertools.count():
            try:
                device_information = self.get_device_information_by_index(n)
            except IndexError:
                break
            info.append(device_information)
        return info

    def set_switch(self, identifier, on=True):
        """
        Sets a switch state on devices providing a switch state.
        'identifier' must be the AIN of the device. 'on' is a boolean
        whether the switch should be on (True) or off (False).
        This method has no return value.
        Raise a FritzArgumentError on invalid identifier.
        """
        arguments = {
            'NewAIN': identifier,
            'NewSwitchState': 'ON' if on else 'OFF'
        }
        self._action('SetSwitch', arguments=arguments)



"""
Module to read status-information from an AVM FritzBox.
"""

from collections import namedtuple
from warnings import warn

from .fritzbase import AbstractLibraryBase
from .fritztools import (
    ArgumentNamespace,
    format_dB,
    format_num,
    format_rate,
)

DefaultConnectionService = namedtuple(
    "DefaultConnectionService",
    "prefix connection_service postfix"
)

def _integer_or_original(value):
    """
    Tries to convert value to an integer. Returns this integer on
    success, otherwise returns the original value.
    """
    try:
        return int(value)
    except ValueError:
        return value


class FritzStatus(AbstractLibraryBase):
    """
    Class for requesting status-information: up, down, ip, activity
    (bytes per second send/received). All parameters are optional. If
    given, they have the following meaning: `fc` is an instance of
    FritzConnection, `address` the ip of the Fritz!Box, `port` the port
    to connect to, `user` the username, `password` the password,
    `timeout` a timeout as floating point number in seconds, `use_tls` a
    boolean indicating to use TLS (default False).
    """

    @property
    def is_linked(self):
        """
        A boolean whether the FritzBox is physically linked to
        the provider.
        """
        status = self.fc.call_action("WANCommonIFC", "GetCommonLinkProperties")
        return status["NewPhysicalLinkStatus"] == "Up"

    @property
    def is_connected(self):
        """
        A boolean whether the FritzBox has established an
        internet-connection.
        """
        status = self.fc.call_action("WANIPConn", "GetStatusInfo")
        return status["NewConnectionStatus"] == "Connected"

    @property
    def external_ip(self):
        """The external v4 ip-address."""
        return self.fc.call_action("WANIPConn", "GetExternalIPAddress")[
            "NewExternalIPAddress"
        ]

    @property
    def external_ipv6(self):
        """The external v6 ip-address."""
        return self.external_ipv6_info["NewExternalIPv6Address"]

    @property
    def external_ipv6_info(self):
        """
        Returns the ipv6 external address information as a dictionary with the keys:
        NewExternalIPv6Address                   out ->     string
        NewPrefixLength                          out ->     ui1
        NewValidLifetime                         out ->     ui4
        NewPreferedLifetime                      out ->     ui4
        """
        return self.fc.call_action("WANIPConn", "X_AVM_DE_GetExternalIPv6Address")

    @property
    def ipv6_prefix(self):
        """The internal v6 prefix."""
        return self.ipv6_prefix_info["NewIPv6Prefix"]

    @property
    def ipv6_prefix_info(self):
        """
        Returns the ipv6 prefix information as a dictionary with the keys:
        NewIPv6Prefix                            out ->     string
        NewPrefixLength                          out ->     ui1
        NewValidLifetime                         out ->     ui4
        NewPreferedLifetime                      out ->     ui4
        """
        return self.fc.call_action("WANIPConn", "X_AVM_DE_GetIPv6Prefix")

    @property
    def connection_uptime(self):
        """Connection uptime in seconds."""
        status = self.fc.call_action("WANIPConn", "GetStatusInfo")
        return status["NewUptime"]

    @property
    def uptime(self):
        """
        .. deprecated:: 1.9.0
           Use :func:`connection_uptime` instead.
        """
        warn('This method is deprecated. Use "connection_uptime" instead.', DeprecationWarning)
        return self.connection_uptime

    @property
    def device_uptime(self):
        """Device uptime in seconds."""
        status = self.fc.call_action("DeviceInfo1", "GetInfo")
        return status["NewUpTime"]

    @property
    def str_uptime(self):
        """Connection uptime in human-readable format."""
        mins, secs = divmod(self.uptime, 60)
        hours, mins = divmod(mins, 60)
        return "%02d:%02d:%02d" % (hours, mins, secs)

    @property
    def bytes_sent(self):
        """Total number of sent bytes."""
        try:
            status = self.fc.call_action("WANCommonIFC1", "GetAddonInfos")
            value = status["NewX_AVM_DE_TotalBytesSent64"]
        except KeyError:
            # fallback for older FritzOS Versions not providing a 64 bit value:
            status = self.fc.call_action(
                "WANCommonIFC1", "GetTotalBytesSent"
            )
            value = status["NewTotalBytesSent"]
        return _integer_or_original(value)

    @property
    def bytes_received(self):
        """Total number of received bytes."""
        try:
            status = self.fc.call_action("WANCommonIFC1", "GetAddonInfos")
            value = status["NewX_AVM_DE_TotalBytesReceived64"]
        except KeyError:
            # fallback for older FritzOS Versions not providing a 64 bit value:
            status = self.fc.call_action(
                "WANCommonIFC1", "GetTotalBytesReceived"
            )
            value = status["NewTotalBytesReceived"]
        return _integer_or_original(value)

    @property
    def transmission_rate(self):
        """
        The upstream and downstream values as a tuple in bytes per
        second. Use this for periodical calling.
        """
        status = self.fc.call_action("WANCommonIFC1", "GetAddonInfos")
        upstream = status["NewByteSendRate"]
        downstream = status["NewByteReceiveRate"]
        return upstream, downstream

    @property
    def str_transmission_rate(self):
        """
        Tuple of human-readable transmission rate in bytes. First item
        is upstream, second item downstream.
        """
        upstream, downstream = self.transmission_rate
        return (format_num(upstream), format_num(downstream))

    @property
    def max_linked_bit_rate(self):
        """
        Tuple with the maximum upstream- and downstream-rate
        of the physical link. The rate is given in bits/sec.
        """
        return self._get_max_bit_rate("WANCommonInterfaceConfig")

    @property
    def max_bit_rate(self):
        """
        Tuple with the maximum upstream- and downstream-rate
        of the given connection. The rate is given in bits/sec.
        """
        return self._get_max_bit_rate("WANCommonIFC")

    def _get_max_bit_rate(self, servicename):
        """
        internal method to get the upstream and downstream-rates for
        different services of the WANCommonInterfaceConfig1 ServiceType.
        """
        status = self.fc.call_action(servicename, "GetCommonLinkProperties")
        downstream = status["NewLayer1DownstreamMaxBitRate"]
        upstream = status["NewLayer1UpstreamMaxBitRate"]
        return upstream, downstream

    @property
    def max_byte_rate(self):
        """
        Same as max_bit_rate but rate is given in bytes/sec.
        """
        upstream, downstream = self.max_bit_rate
        return upstream / 8.0, downstream / 8.0

    @property
    def str_max_linked_bit_rate(self):
        """
        Human-readable maximum of the physical upstream- and
        downstream-rate in bits/sec. Value is a tuple, first item is
        upstream, second item is downstream.
        """
        upstream, downstream = self.max_linked_bit_rate
        return (
            format_rate(upstream, unit="bits"),
            format_rate(downstream, unit="bits"),
        )

    @property
    def str_max_bit_rate(self):
        """
        Human-readable maximum of the upstream- and downstream-rate in
        bits/sec, as given by the provider. Value is a tuple, first item
        is upstream, second item is downstream.
        """
        upstream, downstream = self.max_bit_rate
        return (
            format_rate(upstream, unit="bits"),
            format_rate(downstream, unit="bits"),
        )

    def get_monitor_data(self, sync_group_index=0):
        """
        Returns a dictionary with realtime data about the current up-
        and downstream rates.
        """
        monitor_data = self.fc.call_action(
            "WANCommonInterfaceConfig1",
            "X_AVM-DE_GetOnlineMonitor",
            NewSyncGroupIndex=sync_group_index,
        )
        for key, value in monitor_data.items():
            if isinstance(value, str) and "," in value:
                try:
                    items = [int(v) for v in value.split(",")]
                except (AttributeError, ValueError):
                    # ignore and keep value as is:
                    pass
                else:
                    monitor_data[key] = items
        return monitor_data

    def reconnect(self):
        """Makes a reconnection with a new external ip."""
        self.fc.reconnect()

    @property
    def noise_margin(self):
        """
        Tuple of noise margin. First item
        is upstream, second item downstream.
        """
        status = self.fc.call_action("WANDSLInterfaceConfig1", "GetInfo")
        upstream = status["NewUpstreamNoiseMargin"]
        downstream = status["NewDownstreamNoiseMargin"]
        return upstream, downstream

    @property
    def str_noise_margin(self):
        """
        Human-readable noise margin in dB. Value is a tuple, first item
        is upstream, second item downstream.
        """
        upstream, downstream = self.noise_margin
        return (format_dB(upstream), format_dB(downstream))

    @property
    def attenuation(self):
        """
        Tuple of attenuation. First item
        is upstream, second item downstream.
        """
        status = self.fc.call_action("WANDSLInterfaceConfig1", "GetInfo")
        upstream = status["NewUpstreamAttenuation"]
        downstream = status["NewDownstreamAttenuation"]
        return upstream, downstream

    @property
    def str_attenuation(self):
        """
        Human-readable attenuation in dB. Value is a tuple, first item
        is upstream, second item downstream.
        """
        upstream, downstream = self.attenuation
        return (format_dB(upstream), format_dB(downstream))

    @property
    def upnp_enabled(self):
        """
        Returns a boolean whether upnp is enabled or raises a
        FritzServiceError in case the service is not available.
        """
        status = self.fc.call_action("X_AVM-DE_UPnP1", "GetInfo")
        return status["NewEnable"]

    @property
    def device_has_mesh_support(self):
        """
        True if the device supports mesh, otherwise False.
        """
        # check for the corresponding action
        # whether mesh is supported
        try:
            return (
                "X_AVM-DE_GetMeshListPath"
                in self.fc.services["Hosts1"].actions
            )
        except KeyError:
            # can happen if "Hosts1" is not known
            return False

    def get_device_info(self):
        """
        Returns an ArgumentNamespace with the attributes:

        manufacturer_name, manufacturer_oui, model_name, description,
        product_class, serial_number, software_version, hardware_version,
        spec_version, provisioning_code, up_time, device_log

        .. versionadded:: 1.10

        """
        return ArgumentNamespace(fc.call_action("DeviceInfo1", "GetInfo"))

    def get_default_connection_service(self):
        """
        Returns a namedtuple of type DefaultConnectionService:
        `prefix` -> str
        `device_connection` -> str (like "WANPPPConnection")
        `postfix` -> str
        """
        result = self.fc.call_action(
                "Layer3Forwarding1", "GetDefaultConnectionService"
        )
        prefix, connection_service, postfix = \
            result["NewDefaultConnectionService"].split('.', 2)
        return DefaultConnectionService(
            prefix, connection_service, postfix
        )

    @property
    def connection_service(self):
        """
        The extracted connection_service from
        get_default_connection_service().
        """
        result = self.get_default_connection_service()
        return result.connection_service

    @property
    def update_available(self):
        """
        The new version number (as a string) if an update is available or an
        empty string if no update is avilable.
        """
        return self.fc.call_action("UserInterface1", "GetInfo")["NewX_AVM-DE_Version"]

    @property
    def has_wan_enabled(self):
        """
        True if wan is enabled otherwise False.
        """
        return self.fc.call_action(self.connection_service, "GetInfo")["NewEnable"]

    @property
    def has_wan_support(self):
        """
        True is the device supports a WAN interface.
        False otherwise.
        """
        return "Layer3Forwarding1" in self.fc.services



"""
fritzstatus.py

Module to inspect the FritzBox API for available services and actions.
CLI interface.

This module is part of the FritzConnection package.
https://github.com/kbr/fritzconnection
License: MIT (https://opensource.org/licenses/MIT)
Author: Klaus Bremer
"""

from ..core.exceptions import FritzServiceError, FritzActionError
from ..lib.fritzstatus import FritzStatus
from .utils import get_cli_arguments, get_instance, print_header


def print_status(fs):
    print("FritzStatus:\n")
    status_information = [
        ("is linked", "is_linked"),
        ("is connected", "is_connected"),
        ("external ip (v4)", "external_ip"),
        ("external ip (v6)", "external_ipv6"),
        ("internal ipv6-prefix", "ipv6_prefix"),
        ("uptime", "str_uptime"),
        ("bytes send", "bytes_sent"),
        ("bytes received", "bytes_received"),
        ("max. bit rate", "str_max_bit_rate"),
    ]
    for status, attribute in status_information:
        try:
            information = getattr(fs, attribute)
        except (FritzServiceError, FritzActionError):
            information = f'unsupported attribute "{attribute}"'
        print(f"    {status:22}: {information}")
    print()


def main():
    args = get_cli_arguments()
    if not args.password:
        print("Exit: password required.")
    else:
        fs = get_instance(FritzStatus, args)
        print_header(fs)
        print_status(fs)


if __name__ == "__main__":
    main()



# fritzconnection library:
#
# The modules in the lib folder are helper modules to access some services
# of the FritzBox. They can also be used as examples on how to write
# applications using fritzconnection.



"""
fritzwlan.py

Module to inspect the FritzBox API for wlan devices.
CLI interface.

This module is part of the FritzConnection package.
https://github.com/kbr/fritzconnection
License: MIT (https://opensource.org/licenses/MIT)
Author: Klaus Bremer
"""

import itertools

from ..core.exceptions import FritzServiceError
from ..lib.fritzwlan import FritzWLAN, SERVICE
from . utils import get_cli_arguments, get_instance, print_header


def get_header():
    index = 'index'
    status = 'active'
    mac = 'mac'
    ip = 'ip'
    signal = 'signal'
    speed = 'speed'
    return f'{index:>5}{status:>8}{mac:>20}{ip:>18}{signal:>8}{speed:>8}'


def report_wlanconfiguration(fw, extension):
    fw.service = extension
    hosts_info = fw.get_hosts_info()
    if hosts_info:
        print(f'Hosts registered at {SERVICE}{extension}:')
        print(f'WLAN name: {fw.ssid}')
        print(f'channel  : {fw.channel}')
        print(get_header())
        for info in hosts_info:
            index = info['index']
            status = info['status']
            mac = info['mac']
            ip = info['ip']
            signal = info['signal']
            speed = info['speed']
            print(f'{index:>5}{status:>8}{mac:>20}{ip:>18}{signal:>8}{speed:>8}')
        print()


def report_devices(fw, args):
    if args.service:
        try:
            report_wlanconfiguration(fw, args.service)
        except FritzServiceError as err:
            print(f'Error: {err}')
    else:
        for n in itertools.count(1):
            try:
                report_wlanconfiguration(fw, n)
            except FritzServiceError:
                break


def add_arguments(parser):
    parser.add_argument('-s', '--service',
                        nargs='?', default=0, const=None,
                        help='WLANConfiguration service number')


def main():
    args = get_cli_arguments(add_arguments)
    if not args.password:
        print('Exit: password required.')
    else:
        fw = get_instance(FritzWLAN, args)
        print_header(fw)
        report_devices(fw, args)


if __name__ == '__main__':
    main()



"""
fritzphonebook.py

Module to inspect the Fritz!Box phonebooks.
CLI interface.

This module is part of the FritzConnection package.
https://github.com/kbr/fritzconnection
License: MIT (https://opensource.org/licenses/MIT)
Authors: Klaus Bremer, David M. Straub
"""

from ..lib.fritzphonebook import FritzPhonebook
from . utils import get_cli_arguments, get_instance, print_header


def print_phonebooks(fpb):
    for id in fpb.phonebook_ids:
        info = fpb.phonebook_info(id)
        print(f"Content of phonebook: {info['name']} ")
        for name, numbers in fpb.get_all_names(id).items():
            print(f"{name:<30}{', '.join(numbers)}")
        print()


def print_search_name(fpb, arguments):
    found = False
    for id in fpb.phonebook_ids:
        contacts = fpb.get_all_names(id)
        numbers = contacts.get(arguments.name)
        if numbers:
            print(f"{arguments.name:<30}{', '.join(numbers)}")
            found = True
    if not found:
        print(f"name {arguments.name} not found.")


def print_search_number(fpb, arguments):
    found = False
    for id in fpb.phonebook_ids:
        numbers = fpb.get_all_numbers(id)
        if arguments.number in numbers:
            print(f"{numbers[arguments.number]:<30}{arguments.number}")
            found = True
    if not found:
        print(f'number {arguments.number} not found.')


def add_arguments(parser):
    parser.add_argument('-a', '--all',
                        action='store_true',
                        help='List all phone books ')
    parser.add_argument('--name',
                        nargs='?', default=0,
                        help='Name for number search')
    parser.add_argument('--number',
                        nargs='?', default=0,
                        help='Number for name search')


def main():
    args = get_cli_arguments(add_arguments)
    if not args.password:
        print('Exit: password required.')
        return
    fpb = get_instance(FritzPhonebook, args)
    print_header(fpb)
    print('FritzPhonebook:\n')
    if args.all:
        print_phonebooks(fpb)
    elif args.name:
        print_search_name(fpb, args)
    elif args.number:
        print_search_number(fpb, args)
    print()  # blank line for better readability


if __name__ == '__main__':
    main()



"""
Module to parse and store the device description and the service data
provided by xml and the scpd protocol based on xml. Names partly violate
PEP8 representing node-names from xml description files.
"""
# This module is part of the FritzConnection package.
# https://github.com/kbr/fritzconnection
# License: MIT (https://opensource.org/licenses/MIT)
# Author: Klaus Bremer


from .utils import (
    get_xml_root,
    localname,
)


# ---------------------------------------------------------
# processor, decorator and descriptors here:
# ---------------------------------------------------------

def process_node(obj, root):
    """
    Take an object and a root of nodes. The node.text of nodes with the
    same name as an instance-attribute of 'obj' are set as values for
    the corresponding instance-attribute. If the attribute is a
    callable, processing the node gets delegated to the callable (which
    in turn calls process_node).
    """
    for node in root:
        node_name = localname(node)
        try:
            attr = getattr(obj, node_name)
        except AttributeError:
            # ignore unknown nodes
            continue
        if callable(attr):
            # attribute is a callable: delegate further
            attr(node)
        else:
            # node is an attribute: set value
            if isinstance(node.text, str):
                value = node.text.strip()
            else:
                value = node.text
            setattr(obj, node_name, value)


def processor(cls):
    """
    Class decorator to add the functionality of calling 'process_node'
    on invoking an instance as a callable.
    """
    cls.__call__ = lambda obj, root: process_node(obj, root)
    return cls


class ValueSequencer:
    """
    Data descriptor storing a value (assigned as attribute value) in a
    given sequence.
    """
    def __init__(self, sequence_name):
        self.sequence_name = sequence_name

    def __get__(self, obj, objtype):
        # kind of set only attribute
        return NotImplemented

    def __set__(self, obj, value):
        sequence = getattr(obj, self.sequence_name)
        sequence.append(value)


class InstanceAttributeFactory:
    """
    Non data descriptor returning instances of 'cls' and registering
    these instances in the '_storage' attribute of the calling instance.
    """
    def __init__(self, cls):
        self.cls = cls

    def __get__(self, obj, objtype):
        instance = self.cls()
        obj._storage.append(instance)
        return instance


class Storage:
    """
    Baseclass for classes working with InstanceAttributeFactory.
    """
    def __init__(self, storage):
        self._storage = storage


# ---------------------------------------------------------
# Processors for reading the device specific FritzOS api:
# Node- and root-classes. Root classes are entry-points.
# Root classes are Scpd and Description.
# Description is the only public class.
# ---------------------------------------------------------


class Serializer:
    """
    Simple abstract Serializer base class.
    This class and should not get instanciated.
    """

    def __eq__(self, other):
        if set(self.__dict__.keys()) ^ set(other.__dict__.keys()):
            # self and other have not the same set of instance attributes:
            return False
        # both instances must have all the same attribute values:
        return self._compare_attributes(other, self.__dict__.keys())

    def _compare_attributes(self, other, attributes):
        # check for the same values in the attributes of self and other:
        for attribute in attributes:
            if getattr(self, attribute) != getattr(other, attribute):
                return False
        return True

    def serialize(self, exclude=None):
        if exclude is None:
            exclude = []
        attribute_names = set(self.__dict__.keys()) - set(exclude)
        return self.get_sorted_dict(
            {name: getattr(self, name) for name in attribute_names}
        )

    def deserialize(self, data):
        self.__dict__.update(data)

    @staticmethod
    def get_sorted_dict(dictionary):
        """
        Takes a dictionary and returns another one with all keys in
        alphabetical order.
        """
        sorted_keys = sorted(dictionary.keys())
        return {key: dictionary[key] for key in sorted_keys}

    @classmethod
    def from_data(cls, data):
        """
        Return a new instance with attributes initialized from data.
        """
        try:
            instance = cls()
        except TypeError:
            # will happen on classes expecting a list of nodes:
            instance = cls(root=[])
        instance.deserialize(data)
        return instance


@processor
class SpecVersion(Serializer):
    """
    Specification version from the schema device or service
    information.
    """
    def __init__(self):
        # attributes are case sensitive node names:
        self.major = None
        self.minor = None

    @property
    def version(self):
        return f'{self.major}.{self.minor}'


@processor
class SystemVersion(Serializer):
    """
    Information about the Fritz!OS version of the Fritz!Box.
    Information is just provided by the 'tr64desc.xml' file.
    """
    def __init__(self):
        # attributes are case sensitive node names
        self.HW = None
        self.Major = None
        self.Minor = None
        self.Patch = None
        self.Buildnumber = None
        self.Display = None

    @property
    def version(self):
        """
        Returns system version as string like '7.10' or None if system
        version is unknown.
        """
        if self.Minor and self.Patch:
            return f'{self.Minor}.{self.Patch}'
        return None

    @property
    def info(self):
        """
        Returns a tuple with all instance attributes 'HW, Major, Minor,
        Patch, Buildnumber, Display' in this order.
        """
        return (
            self.HW,
            self.Major,
            self.Minor,
            self.Patch,
            self.Buildnumber,
            self.Display,
        )


@processor
class Argument(Serializer):
    """
    An argument with name, direction and relatedStateVariable
    attributes.
    """
    def __init__(self):
        # attributes are case sensitive node names
        self.name = None
        self.direction = None
        self.relatedStateVariable = None


@processor
class ArgumentList(Storage):
    """
    Collects the arguments for an action.
    """
    # case sensitive node
    argument = InstanceAttributeFactory(Argument)


@processor
class Action(Serializer):
    """
    Every Action has a name and a list of arguments.
    """
    def __init__(self):
        self._arguments = list()
        self._arguments_storage = None
        # attributes are case sensitive node names:
        self.name = None
        self.argumentList = ArgumentList(self._arguments)

    def __eq__(self, other):
        # for testing: Action is equal to another if the name is the same
        # and the arguments in self._arguments are the same and in the
        # same order (because of the implementation).
        if self.name == other.name:
            for arg1, arg2 in zip(self._arguments, other._arguments):
                if arg1 != arg2:
                    return False
            return True
        return False

    @property
    def arguments(self):
        """
        Returns the action-arguments as a dict. argument-names are the
        keys and the argument objects are the values. The dictionary
        gets cached.
        """
        if not self._arguments_storage:
            self._arguments_storage = {arg.name: arg for arg in self._arguments}
        return self._arguments_storage

    def serialize(self):
        """
        Return a dictionary with json serializable data: the name of the
        instance and a list of the serialized argument-instances in
        self._arguments.
        """
        exclude = ["_arguments", "_arguments_storage", "argumentList"]
        data = super().serialize(exclude=exclude)
        data['arguments'] = [arg.serialize() for arg in self._arguments]
        return data

    def deserialize(self, data):
        """
        Deserialize the data back to an Action instance with a given
        name and defined Argument-instances.
        """
        self.name = data['name']
        self._arguments = [Argument.from_data(d) for d in data['arguments']]


@processor
class ActionList(Storage):
    """
    Collection of actions of a service.
    The Action instances are stored in the Scpd.actions attribute.
    """
    # case sensitive node
    action = InstanceAttributeFactory(Action)


@processor
class ValueRange(Serializer):

    def __init__(self):
        # attributes are case sensitive node names:
        self.minimum = None
        self.maximum = None
        self.step = None


@processor
class StateVariable(Serializer):
    """
    Represents a stateVariable with the attributes name, dataType,
    defaultValue, allowedValueList and allowedValueRange.
    """
    # case sensitive node
    allowedValue = ValueSequencer('allowed_values')

    def __init__(self):
        # attributes are case sensitive node names:
        self.name = None
        self.dataType = None
        self.defaultValue = None
        self.allowed_values = list()  # list of values as strings
        self.allowedValueList = self
        self.allowedValueRange = ValueRange()

    def __eq__(self, other):
        # two instances are equal on having the same attribute values.
        attributes = ["name", "dataType", "defaultValue", "allowed_values"]
        if not self._compare_attributes(other, attributes):
            return False
        return self.allowedValueRange == other.allowedValueRange

    def serialize(self):
        """
        Returns a dictionary with json serializable attribute data.
        """
        exclude = ["allowedValueList", "allowedValueRange"]
        data = {"attributes": super().serialize(exclude=exclude)}
        data["allowedValueRange"] = self.allowedValueRange.serialize()
        return data

    def deserialize(self, data):
        """
        Deserialize the data back to a former state of a StateVariable
        instance.
        """
        super().deserialize(data['attributes'])
        self.allowedValueRange.deserialize(data['allowedValueRange'])


@processor
class ServiceStateTable(Storage):
    """
    Collection of stateVariables.
    """
    # case sensitive node
    stateVariable = InstanceAttributeFactory(StateVariable)


class Scpd(Serializer):
    """
    Provides information about the Service Control Point Definitions
    for every Service. Every Service has one instance of this class for
    accessing the description of its own actions and the according
    parameters.
    Root class for processing the content of an scpd-file.
    """
    def __init__(self, root):
        """
        Starts interpreting the scpd-data. 'root' must be a xml.Element
        objects as returned from 'utils.get_xml_root'.
        """
        self._actions = list()
        self._state_variables = list()

        # attributes are case sensitive node names:
        self.specVersion = SpecVersion()
        self.actionList = ActionList(self._actions)
        self.serviceStateTable = ServiceStateTable(self._state_variables)

        # start node processing:
        process_node(self, root)

    def __eq__(self, other):
        attributes = ["_actions", "_state_variables", "specVersion"]
        return self._compare_attributes(other, attributes)

    @property
    def spec_version(self):
        return self.specVersion.version

    @property
    def actions(self):
        """
        Returns a dictionary with the actions from the actions-list. The
        action-names are the keys and the actions themselves are the
        values.
        """
        return {action.name: action for action in self._actions}

    @property
    def state_variables(self):
        """
        Returns a dictionary with the state_variable name as keys and
        the StateVariable itself as value.
        """
        return {sv.name: sv for sv in self._state_variables}

    def serialize(self):
        data = {"actions": [action.serialize() for action in self._actions]}
        data['state_variables'] = [sv.serialize() for sv in self._state_variables]
        data['specVersion'] = self.specVersion.serialize()
        return self.get_sorted_dict(data)

    def deserialize(self, data):
        self._actions = [Action.from_data(d) for d in data['actions']]
        self._state_variables = [StateVariable.from_data(d) for d in data['state_variables']]
        self.specVersion.deserialize(data['specVersion'])


@processor
class Service(Serializer):
    """
    Class describing a service.
    """
    def __init__(self):
        self._scpd = None
        self._actions = None
        self._state_variables = None
        # attributes are case sensitive node names:
        self.serviceType = None
        self.serviceId = None
        self.controlURL = None
        self.eventSubURL = None
        self.SCPDURL = None

    @property
    def name(self):
        if self.serviceId:
            return self.serviceId.split(':')[-1]
        return None

    @property
    def actions(self):
        """
        Returns all known actions of this service as a dictionary.
        Action names are keys, the action objects are the values. Caches
        the dictionary once retrieved from _scpd.
        """
        if self._actions is None:
            self._actions = self._scpd.actions
        return self._actions

    @property
    def state_variables(self):
        """
        Returns all known stateVariables of this service as a
        dictionary. Names are keys, the stateVariables objects are the
        values. Caches the dictionary once retrieved from _scpd.
        """
        if self._state_variables is None:
            self._state_variables = self._scpd.state_variables
        return self._state_variables

    def load_scpd(self, address, port, timeout=None, session=None):
        """Loads the scpd data"""
        url = f'{address}:{port}{self.SCPDURL}'
        root = get_xml_root(url, timeout=timeout, session=session)
        self._scpd = Scpd(root)

    def serialize(self):
        """
        Serialize the service instance attributes. Returns a dictionary
        with data that can be converted to json.
        """
        exclude = ['_scpd', '_actions', '_state_variables']
        return {
            "attributes": super().serialize(exclude=exclude),
            "scpd": self._scpd.serialize()
        }

    def deserialize(self, data):
        """
        Inverse method for `serialize`. Takes the data (a dictionary)
        and populates the instance attributes extracted by `serialize`.
        """
        super().deserialize(data['attributes'])
        self._scpd = Scpd.from_data(data['scpd'])


@processor
class ServiceList(Storage):
    """
    Collection of Service instances for a device.
    The service instances are stored in the device.services attribute.
    """
    # case sensitive node
    service = InstanceAttributeFactory(Service)


@processor
class Device(Serializer):
    """
    Storage for devices attributes and device sub-nodes.
    Sub-nodes are the serviceList and the deviceList.
    The services provided by a device are collected in services.
    Sub-devices are collected in devices.
    All instance attributes are public for read only use.
    """
    def __init__(self):
        self._services = list()
        self.devices = list()
        # attributes are case sensitive node names:
        self.deviceType = None
        self.friendlyName = None
        self.manufacturer = None
        self.manufacturerURL = None
        self.modelDescription = None
        self.modelName = None
        self.modelNumber = None
        self.modelURL = None
        self.UDN = None
        self.UPC = None
        self.presentationURL = None
        self.serviceList = ServiceList(self._services)
        self.deviceList = DeviceList(self.devices)

    def __eq__(self, other):
        attributes = [
            "deviceType", "friendlyName", "manufacturer", "manufacturerURL",
            "modelDescription", "modelName", "modelNumber", "modelURL",
            "UDN", "UPC", "presentationURL",
        ]
        return self._compare_attributes(other, attributes)

    @property
    def services(self):
        services = {service.name: service for service in self._services}
        for device in self.devices:
            services.update(device.services)
        return services

    def serialize(self):
        """
        Returns a dictionary with a subset of the instance attributes
        and a list of serialized services that can be transformed to
        json-format.
        """
        exclude = ["_services", "devices", "serviceList", "deviceList",]
        data = {'attributes': super().serialize(exclude=exclude)}
        data['services'] = [service.serialize() for service in self._services]
        data['devices'] = [device.serialize() for device in self.devices]
        return self.get_sorted_dict(data)

    def deserialize(self, data):
        """
        Loads the data into the instance attributes. This is the
        reverse-function for serialize. No return value.
        """
        super().deserialize(data['attributes'])
        self._services.extend([Service.from_data(d) for d in data['services']])
        self.devices.extend([Device.from_data(d) for d in data['devices']])


@processor
class DeviceList(Storage):
    """
    Collection of sub-devices of a device.
    The Device instances are stored in the device.devices attribute of
    the parent device.
    """
    # case sensitive node
    device = InstanceAttributeFactory(Device)


class Description(Serializer):
    """
    Root class for a given description information as the content from
    the files igddesc.xml or tr64desc.xml.
    """
    def __init__(self, root):
        """
        Starts data-processing. 'root' must be a xml.Element object as
        returned from 'utils.get_xml_root'.
        """
        # attributes are case sensitive node names:
        self.device = Device()
        self.specVersion = SpecVersion()
        self.systemVersion = SystemVersion()

        # start node processing:
        process_node(self, root)

    @property
    def device_model_name(self):
        return self.device.modelName

    @property
    def spec_version(self):
        return self.specVersion.version

    @property
    def system_version(self):
        """
        Returns the system version of the Fritz!Box as a string like
        '7.10' or None. This information is only available by the
        'tr64desc.xml' file.
        """
        return self.systemVersion.version

    @property
    def system_info(self):
        """
        Returns the systemVersion attributes as a tuple:
        (HW, Major, Minor, Patch, Buildnumber, Display). This information
        is only available by the 'tr64desc.xml' file.
        """
        return self.systemVersion.info

    @property
    def system_buildnumber(self):
        """
        Returns the buildnumber or None. This information is only
        available by the 'tr64desc.xml' file.
        """
        return self.systemVersion.Buildnumber

    @property
    def system_display(self):
        """
        Returns the system display-string or None. This information is
        only available by the 'tr64desc.xml' file.
        """
        return self.systemVersion.Display

    @property
    def services(self):
        """
        Returns dictionary with the known services as values and the
        according service-names as keys.
        """
        return self.device.services

    def serialize(self):
        """
        Return serialized instance attributes as dictionary.
        """
        return {
            'device': self.device.serialize(),
            'specVersion': self.specVersion.serialize(),
            'systemVersion': self.systemVersion.serialize(),
        }

    def deserialize(self, data):
        """
        Sets the instance attributes according to data.
        """
        self.device.deserialize(data['device'])
        self.specVersion.deserialize(data['specVersion'])
        self.systemVersion.deserialize(data['systemVersion'])


# ---------------------------------------------------------
# Processors to interpret the information returned from
# the "X_AVM-DE_GetHostListPath" action.
# ---------------------------------------------------------

@processor
class Host:
    """
    Host class providing every requested attribute
    """
    _int_values = {'Index', 'X_AVM-DE_Port', 'X_AVM-DE_Speed'}
    _bool_values = {
        'Active',
        'X_AVM-DE_UpdateAvailable',
        'X_AVM-DE_Guest', 'X_AVM-DE_VPN',
        'X_AVM-DE_Disallow',
    }

    def __getattr__(self, attr_name):
        # do the magic of not raising an AttributeError:
        setattr(self, attr_name, None)
        return getattr(self, attr_name)

    @property
    def attributes(self):
        """
        Provide all attributes of the instance as a dictionary with the
        attribute names as keys and the values converted to python
        datatypes.
        """
        attrs = {}
        for name, value in self.__dict__.items():
            if name in self._int_values:
                attrs[name] = int(value)
            elif name in self._bool_values:
                attrs[name] = bool(int(value))
            else:
                attrs[name] = value
        return attrs


@processor
class HostStorage(Storage):
    """
    Storage class collection all Item-nodes describing the hosts.
    The Item-nodes are converted to _Host instances.
    """
    Item = InstanceAttributeFactory(Host)  # 'Item' must match node-name

    def __init__(self, root):
        self._hosts = list()
        super().__init__(self._hosts)
        self(root)  # start process_node()

    @property
    def hosts_attributes(self):
        """
        Provide a list of dictionaries with the attributes of all hosts.
        The list is sorted with the lowest Index number first.
        """
        # list is already sorted from FritzOS,
        # but don't trust this for any time in the future.
        return sorted(
            [host.attributes for host in self._hosts],
            key = lambda attrs: attrs["Index"]
        )



"""
fritzhomeauto.py

Module to inspect the FritzBox homeautomation API.
CLI interface.

This module is part of the FritzConnection package.
https://github.com/kbr/fritzconnection
License: MIT (https://opensource.org/licenses/MIT)
Author: Klaus Bremer
"""

from ..lib.fritzhomeauto import FritzHomeAutomation
from . utils import get_cli_arguments, get_instance, print_header


def report_verbose(fh):
    information = fh.device_information()
    for info in information:
        width = len(max(info.keys(), key=lambda x: len(x)))
        line = f'{{attribute:{width}}} : {{value}}'
        for attribute in sorted(info.keys()):
            print(line.format(attribute=attribute, value=info[attribute]))
        print()  # add blank line between devices


def report_compact(fh):
    name = 'Device Name'
    ain = 'AIN'
    power = 'Power[W]'
    temperature = 't[°C]'
    switch_state = 'switch'
    print(f'{name:24}{ain:18}{power:>10}{temperature:>8}   {switch_state}')
    for di in fh.device_information():
        name = di['NewDeviceName']
        ain = di['NewAIN']
        ain = f"'{ain}'"
        power = di['NewMultimeterPower'] * 0.01
        temperature = di['NewTemperatureCelsius'] *0.1
        switch_state = di['NewSwitchState'].lower()
        print(f'{name:24}{ain:18}{power:>10.3f}{temperature:>8.1f}   {switch_state}')
    print()


def report_status(fh, arguments):
    print('FritzHomeautomation:')
    print('Status of registered home-automation devices:\n')
    if arguments.verbose:
        report_verbose(fh)
    else:
        report_compact(fh)


def switch_device(fh, arguments):
    ain = arguments.switch[0]
    state = arguments.switch[1].lower() == 'on'
    fh.set_switch(identifier=ain, on=state)


def add_arguments(parser):
    parser.add_argument('-v', '--verbose',
                        nargs='?', default=False, const=True,
                        help='report in verbose mode')
    parser.add_argument('-s', '--switch',
                        nargs=2,
                        help='set switch state. requires two parameters: '
                             'ain and state [on|off]')


def main():
    arguments = get_cli_arguments(add_arguments)
    if not arguments.password:
        print('Exit: password required.')
        return
    fh = get_instance(FritzHomeAutomation, arguments)
    if arguments.switch:
        switch_device(fh, arguments)
    else:
        print_header(fh)
        report_status(fh, arguments)


if __name__ == '__main__':
    main()



"""
fritzconnection

library to communicate with the AVM Fritz!Box
via Soap and TR-064.
provides also interface for realtime call-monitoring.

for documentation refer:
https://avm.de/service/schnittstellen/
https://fritzconnection.readthedocs.io/
"""

__version__ = "1.10.0"

# import shortcuts
from .core.fritzconnection import FritzConnection
from .core.fritzmonitor import FritzMonitor



"""
fritzhosts.py

Module to inspect the FritzBox API for registered hosts.
CLI interface.

This module is part of the FritzConnection package.
https://github.com/kbr/fritzconnection
License: MIT (https://opensource.org/licenses/MIT)
Author: Klaus Bremer
"""

from ..lib.fritzhosts import FritzHosts
from . utils import get_cli_arguments, get_instance, print_header


def print_status(fh):
    print('FritzHosts:')
    print('List of registered hosts:\n')
    print('{:>3}: {:<16} {:<28} {:<17}   {}\n'.format(
        'n', 'ip', 'name', 'mac', 'status'))
    hosts = fh.get_hosts_info()
    for index, host in enumerate(hosts, start=1):
        status = 'active' if host['status'] else  '-'
        ip = host['ip'] if host['ip'] else '-'
        mac = host['mac'] if host['mac'] else '-'
        hn = host['name']
        print(f'{index:>3}: {ip:<16} {hn:<28} {mac:<17}   {status}')
    print('\n')


def main():
    args = get_cli_arguments()
    if not args.password:
        print('Exit: password required.')
    else:
        fh = get_instance(FritzHosts, args)
        print_header(fh)
        print_status(fh)


if __name__ == '__main__':
    main()



import os
import re
from setuptools import setup, find_packages


def get_long_description():
    with open("README.rst") as file:
        return file.read()


def get_version():
    path = os.path.join("fritzconnection", "__init__.py")
    with open(path) as file:
        content = file.read()
    mo = re.search(r'\n\s*__version__\s*=\s*[\'"]([^\'"]*)[\'"]', content)
    if mo:
        return mo.group(1)
    raise RuntimeError(f"Unable to find version string in {path}")


setup(
    name="fritzconnection",
    version=get_version(),
    packages=find_packages(exclude=["*.tests"]),
    license="MIT",
    description="Communicate with the AVM FRITZ!Box",
    long_description=get_long_description(),
    author="Klaus Bremer",
    author_email="bremer@bremer-media.com",
    url="https://github.com/kbr/fritzconnection",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    keywords="AVM FRITZ!Box fritzbox fritz",
    python_requires=">= 3.6",
    install_requires=["requests>=2.22.0",],
    extras_require={
        "qr": ["segno>=1.4.1",],
    },
    entry_points={
        "console_scripts": [
            "fritzconnection = fritzconnection.cli.fritzinspection:main",
            "fritzcall = fritzconnection.cli.fritzcall:main",
            "fritzhomeauto = fritzconnection.cli.fritzhomeauto:main",
            "fritzhosts = fritzconnection.cli.fritzhosts:main",
            "fritzmonitor = fritzconnection.cli.fritzmonitor:main",
            "fritzphonebook = fritzconnection.cli.fritzphonebook:main",
            "fritzstatus = fritzconnection.cli.fritzstatus:main",
            "fritzwlan = fritzconnection.cli.fritzwlan:main",
        ]
    },
)
